"""Tests for DataContainer

Michel Bierlaire
Tue Mar 25 17:43:42 2025
"""

import unittest

import pandas as pd


from biogeme.database import Database
from biogeme.exceptions import BiogemeError
from biogeme.expressions import Variable
from biogeme.floating_point import PANDAS_FLOAT


class TestDataContainer(unittest.TestCase):

    def setUp(self):
        self.df = pd.DataFrame(
            {
                'x': [1, 2, 3],
                'y': [4, 5, 6],
            },
            dtype=PANDAS_FLOAT,
        )
        self.container = Database('self', self.df.copy())

    def test_init_valid(self):
        self.assertEqual(self.container.num_rows(), 3)
        self.assertEqual(self.container.num_columns(), 2)

    def test_init_empty(self):
        with self.assertRaises(BiogemeError):
            Database('test', pd.DataFrame())

    def test_get_column_valid(self):
        col = self.container.get_column('x')
        pd.testing.assert_series_equal(
            col.reset_index(drop=True),
            pd.Series([1, 2, 3], name='x', dtype=PANDAS_FLOAT),
        )

    def test_get_column_invalid(self):
        with self.assertRaises(BiogemeError):
            self.container.get_column('z')

    def test_column_exists(self):
        self.assertTrue(self.container.column_exists('x'))
        self.assertFalse(self.container.column_exists('z'))

    def test_add_column_success(self):
        new_col = pd.Series([7, 8, 9])
        self.container.add_column('z', new_col)
        self.assertTrue(self.container.column_exists('z'))

    def test_add_column_duplicate(self):
        with self.assertRaises(ValueError):
            self.container.add_column('x', pd.Series([1, 2, 3]))

    def test_add_column_wrong_length(self):
        with self.assertRaises(ValueError):
            self.container.add_column('z', pd.Series([1, 2]))

    def test_scale_column_success(self):
        self.container.scale_column('x', 2)
        expected = pd.Series([2, 4, 6], name='x', dtype=PANDAS_FLOAT)
        pd.testing.assert_series_equal(self.container.get_column('x'), expected)

    def test_scale_column_missing(self):
        with self.assertRaises(BiogemeError):
            self.container.scale_column('z', 2)

    def test_remove_column(self):
        self.container.remove_column('x')
        self.assertFalse(self.container.column_exists('x'))

    def test_remove_rows(self):
        condition = self.df['x'] > 1
        self.container.remove_rows(condition)
        self.assertEqual(self.container.num_rows(), 1)
        pd.testing.assert_series_equal(
            self.container.get_column('x'), pd.Series([1], name='x', dtype=PANDAS_FLOAT)
        )

    def test_remove_with_expression(self):

        x = Variable('x')
        condition = x > 1
        self.container.remove(condition)
        self.assertEqual(self.container.num_rows(), 1)
        pd.testing.assert_series_equal(
            self.container.get_column('x'), pd.Series([1], name='x', dtype=PANDAS_FLOAT)
        )

    def test_dataframe_property(self):
        df = self.container.dataframe
        self.assertIsInstance(df, pd.DataFrame)
        self.assertEqual(list(df.columns), ['x', 'y'])

    def test_define_variable(self):
        x = Variable('x')
        expr = x * 10
        self.container.define_variable('x_scaled', expr)
        expected = pd.Series([10.0, 20.0, 30.0], name='x_scaled', dtype=PANDAS_FLOAT)
        pd.testing.assert_series_equal(self.container.get_column('x_scaled'), expected)

    def test_dataframe_dtype(self):
        # Check that all columns have the correct dtype
        for col in self.container.dataframe.columns:
            self.assertEqual(self.container.dataframe[col].dtype, PANDAS_FLOAT)

    def test_bootstrap_sample_dtype(self):
        # Check that the bootstrap sample also uses the correct float type
        bootstrap = self.container.bootstrap_sample()
        for col in bootstrap.dataframe.columns:
            self.assertEqual(bootstrap.dataframe[col].dtype, PANDAS_FLOAT)

    def test_define_variable_dtype(self):
        # Define a new variable and check its dtype
        x = Variable('x')
        self.container.define_variable('x_squared', x * x)
        self.assertEqual(self.container.get_column('x_squared').dtype, PANDAS_FLOAT)

    def test_scale_column_dtype(self):
        # Scale a column and verify dtype remains correct
        self.container.scale_column('x', 1.5)
        self.assertEqual(self.container.get_column('x').dtype, PANDAS_FLOAT)

    def test_add_column_dtype(self):
        # Add a column and check its dtype
        new_col = pd.Series([0.1, 0.2, 0.3], dtype=PANDAS_FLOAT)
        self.container.add_column('z', new_col)
        self.assertEqual(self.container.get_column('z').dtype, PANDAS_FLOAT)


if __name__ == '__main__':
    unittest.main()



"""Tests for deep and flat copy

Michel Bierlaire
Fri Jun 27 2025, 17:30:22
"""

import unittest

import pandas as pd
from biogeme.catalog import Catalog, CentralController, segmentation_catalogs
from biogeme.database import Database
from biogeme.expressions import (
    BelongsTo,
    Beta,
    BinaryMax,
    BinaryMin,
    Numeric,
    Variable,
    log,
)
from biogeme.models import loglogit


class TestExpressions(unittest.TestCase):
    def test_belongs_to(self):
        the_set = {1, 2, 3}
        value = Numeric(1)
        expression = BelongsTo(child=value, the_set=the_set)
        copy_expression = expression.deep_flat_copy()
        self.assertEqual(expression.the_set, copy_expression.the_set)
        self.assertEqual(
            expression.child.get_value(), copy_expression.child.get_value()
        )

    def test_belongs_to_with_numeric(self):
        the_set = {1, 2, 3}
        expression = BelongsTo(child=1, the_set=the_set)
        copy_expression = expression.deep_flat_copy()
        self.assertEqual(expression.the_set, copy_expression.the_set)
        self.assertEqual(
            expression.child.get_value(), copy_expression.child.get_value()
        )

    def test_beta(self):
        beta = Beta(name='test', value=0.1, lowerbound=-1, upperbound=1, status=0)
        copy_beta = beta.deep_flat_copy()
        self.assertEqual(beta.name, copy_beta.name)
        self.assertEqual(beta.init_value, copy_beta.init_value)
        self.assertEqual(beta.lower_bound, copy_beta.lower_bound)
        self.assertEqual(beta.upper_bound, copy_beta.upper_bound)
        self.assertEqual(beta.status, copy_beta.status)

    def test_max(self):
        expression = BinaryMax(1, 2)
        copy_expression = expression.deep_flat_copy()
        self.assertEqual(expression.left.get_value(), copy_expression.left.get_value())
        self.assertEqual(
            expression.right.get_value(), copy_expression.right.get_value()
        )

    def test_min(self):
        expression = BinaryMin(1, 2)
        copy_expression = expression.deep_flat_copy()
        self.assertEqual(expression.left.get_value(), copy_expression.left.get_value())
        self.assertEqual(
            expression.right.get_value(), copy_expression.right.get_value()
        )


class TestMultipleExpressions(unittest.TestCase):
    def setUp(self):
        data = {
            'choice': [0, 1, 2, 1, 0],
            'male': [0, 1, 0, 1, 0],
            'train_tt': [15.593820, 11.937003, 8.916655, 1.999498, 9.184978],
            'train_cost': [6.674172, 2.857336, 13.017769, 1.128232, 14.439975],
            'sm_tt': [18.771054, 0.015575, 19.844231, 12.349630, 12.233063],
            'sm_cost': [0.141326, 0.461249, 10.495493, 7.997219, 0.933313],
            'car_tt': [19.475110, 4.655427, 1.812129, 12.367720, 7.649240],
            'car_cost': [19.664618, 9.335258, 17.198808, 13.606151, 9.009985],
        }

        df = pd.DataFrame(data)

        self.database = Database(name='test', dataframe=df)
        asc_car = Beta('asc_car', 0, None, None, 0)
        asc_train = Beta('asc_train', 0, None, None, 0)
        b_time = Beta('b_time', 0, None, None, 0)
        b_cost = Beta('b_cost', 0, None, None, 0)

        # %%
        segmentation_gender = self.database.generate_segmentation(
            variable=Variable('male'), mapping={0: 'female', 1: 'male'}
        )

        # %%
        # We define catalogs with two different specifications for the
        # ASC_CAR: non segmented, and segmented.
        asc_train_catalog, asc_car_catalog = segmentation_catalogs(
            generic_name='asc',
            beta_parameters=[asc_train, asc_car],
            potential_segmentations=(segmentation_gender,),
            maximum_number=1,
        )

        # %%
        # We now define a catalog  with the log travel time as well as the travel time.

        # %%
        # First for train
        train_tt_catalog = Catalog.from_dict(
            catalog_name='train_tt_catalog',
            dict_of_expressions={
                'linear': Variable('train_tt'),
                'log': log(Variable('train_tt')),
            },
        )
        # %%
        # Then for SM. But we require that the specification is the same as
        # train by defining the same controller.
        sm_tt_catalog = Catalog.from_dict(
            catalog_name='sm_tt_catalog',
            dict_of_expressions={
                'linear': Variable('sm_tt'),
                'log': Variable('sm_tt'),
            },
            controlled_by=train_tt_catalog.controlled_by,
        )

        # %%
        # Definition of the utility functions with linear cost.
        v_train = (
            asc_train_catalog
            + b_time * train_tt_catalog
            + b_cost * Variable('train_cost')
        )
        v_swissmetro = b_time * sm_tt_catalog + b_cost * Variable('sm_cost')
        v_car = (
            asc_car_catalog
            + b_time * Variable('car_tt')
            + b_cost * Variable('car_cost')
        )

        # %%
        # Associate utility functions with the numbering of alternatives.
        v = {1: v_train, 2: v_swissmetro, 3: v_car}

        # %%
        # Definition of the model. This is the contribution of each
        # observation to the log likelihood function.
        self.log_probability = loglogit(v, None, Variable('choice'))

        central_controller = CentralController(
            expression=self.log_probability, maximum_number_of_configurations=10
        )
        self.the_iterator = central_controller.expression_iterator()

    def test_multiple_expressions(self):
        self.assertTrue(self.log_probability.embed_expression(Catalog))
        for named_expression in self.the_iterator:
            flat_expression = named_expression.expression.deep_flat_copy()
            self.assertFalse(flat_expression.embed_expression(Catalog))



"""Arithmetic expressions accepted by Biogeme: elementary expressions

Michel Bierlaire
Tue Mar 25 17:34:47 2025
"""

from __future__ import annotations

import logging

from .base_expressions import Expression
from .elementary_types import TypeOfElementaryExpression

logger = logging.getLogger(__name__)


class Elementary(Expression):
    """Elementary expression.

    It is typically defined by a name appearing in an expression. It
    can be a variable (from the database), or a parameter (fixed or to
    be estimated using maximum likelihood), a random variable for
    numerical integration, or Monte-Carlo integration.

    """

    expression_type = None

    def __init__(self, name: str):
        """Constructor

        :param name: name of the elementary expression.
        :type name: string

        """
        super().__init__()
        self.name = name  #: name of the elementary expression

        # self.elementary_index = None
        """The index should be unique for all elementary expressions
        appearing in a given set of formulas.
        """
        self.specific_id: int | None = None  # Index of the element in its own array.

    def __str__(self) -> str:
        """string method

        :return: name of the expression
        :rtype: str
        """
        return f"{self.name}"

    def __repr__(self):
        return f'<{self.get_class_name()} name={self.name}>'

    def get_elementary_expression(self, name: str) -> Expression | None:
        """

        :return: an elementary expression from its name if it appears in the
            expression. None otherwise.
        :rtype: biogeme.Expression
        """
        if self.name == name:
            return self

        return None

    def rename_elementary(
        self, old_name: str, new_name: str, elementary_type: TypeOfElementaryExpression
    ) -> int:
        """Rename an elementary expression
        :return: number of modifications actually performed
        """
        if self.expression_type == elementary_type and self.name == old_name:
            self.name = new_name
            return 1
        return 0

    def dict_of_elementary_expression(
        self, the_type: TypeOfElementaryExpression
    ) -> dict[str, Elementary]:
        """Extract a dict with all elementary expressions of a specific type

        :param the_type: the type of expression
        """
        if self.expression_type == the_type:
            return {self.name: self}
        return {}

    def set_specific_id(self, name, specific_id, the_type: TypeOfElementaryExpression):
        """The elementary IDs identify the position of each element in the corresponding datab"""
        if the_type == self.expression_type and name == self.name:
            self.specific_id = specific_id


def get_free_beta_values(the_expression: Expression) -> dict[str, float]:
    free_beta_expressions: dict[str:Elementary] = (
        the_expression.dict_of_elementary_expression(
            the_type=TypeOfElementaryExpression.FREE_BETA
        )
    )
    return {
        the_beta.name: the_beta.get_value()
        for the_beta in free_beta_expressions.values()
    }



import logging

import pandas as pd

from biogeme.expressions import Variable
from .segmentation import DiscreteSegmentationTuple
from ..exceptions import BiogemeError

logger = logging.getLogger(__name__)
"""Logger that controls the output of
        messages to the screen and log file.
        """


def generate_segmentation(
    dataframe: pd.DataFrame,
    variable: Variable | str,
    mapping: dict[int, str] | None = None,
    reference: str | None = None,
) -> DiscreteSegmentationTuple:
    """Generate a segmentation tuple for a variable.

    :param dataframe: data frame.
    :param variable: Variable object or name of the variable
    :param mapping: mapping associating values of the variable to
        names. If incomplete, default names are provided.
    :param reference: name of the reference category. If None, an
        arbitrary category is selected as reference.

    """

    the_variable = variable if isinstance(variable, Variable) else Variable(variable)

    # Check if the variable is in the database.
    if the_variable.name not in dataframe.columns:
        error_msg = f'Unknown the_variable {the_variable.name}'
        raise BiogemeError(error_msg)

    # Extract all unique values from the data base.
    unique_values = set(dataframe[the_variable.name].unique())

    if len(unique_values) >= 10:
        warning_msg = (
            f'Variable {the_variable.name} takes a total of '
            f'{len(unique_values)} different values in the database. It is '
            f'likely to be too large for a discrete segmentation.'
        )
        logger.warning(warning_msg)

    # Check that the provided mapping is consistent with the data
    values_not_in_data = [
        value for value in mapping.keys() if value not in unique_values
    ]

    if values_not_in_data:
        error_msg = (
            f'The following values in the mapping do not exist in the data for '
            f'variable {the_variable.name}: {values_not_in_data}'
        )
        raise BiogemeError(error_msg)

    the_mapping = {
        int(value): f'{the_variable.name}_{int(value)}' for value in unique_values
    }

    if mapping is not None:
        the_mapping.update(mapping)

    if reference is not None and reference not in mapping.values():
        error_msg = (
            f'Level {reference} of variable {the_variable.name} does not '
            'appear in the mapping: {mapping.values()}'
        )
        raise BiogemeError(error_msg)

    return DiscreteSegmentationTuple(
        variable=the_variable,
        mapping=the_mapping,
        reference=reference,
    )


def verify_segmentation(
    dataframe: pd.DataFrame, segmentation: DiscreteSegmentationTuple
) -> None:
    """Verifies if the definition of the segmentation is consistent with the data

    :param dataframe: dataframe to check.
    :param segmentation: definition of the segmentation
    :raise BiogemeError: if the segmentation is not consistent with the data.
    """

    variable = segmentation.variable

    # Check if the variable is in the database.
    if variable.name not in dataframe.columns:
        error_msg = f'Unknown variable {variable.name}'
        raise BiogemeError(error_msg)

    # Extract all unique values from the data base.
    unique_values = set(dataframe[variable.name].unique())
    segmentation_values = set(segmentation.mapping.keys())

    in_data_not_in_segmentation = unique_values - segmentation_values
    in_segmentation_not_in_data = segmentation_values - unique_values

    error_msg_1 = (
        (
            f'The following entries are missing in the segmentation: '
            f'{in_data_not_in_segmentation}.'
        )
        if in_data_not_in_segmentation
        else ''
    )

    error_msg_2 = (
        (
            f'Segmentation entries do not exist in the data: '
            f'{in_segmentation_not_in_data}.'
        )
        if in_segmentation_not_in_data
        else ''
    )

    if error_msg_1 or error_msg_2:
        raise BiogemeError(f'{error_msg_1} {error_msg_2}')



"""
This module provides utility functions for performing sampling operations
on pandas DataFrames, including standard bootstrapping and panel-based sampling.

Michel Bierlaire
Wed Mar 26 19:39:21 2025
"""

import pandas as pd
import numpy as np
from biogeme.exceptions import BiogemeError


def sample_with_replacement(df: pd.DataFrame, size: int | None = None) -> pd.DataFrame:
    if size is None:
        size = len(df)
    indices = np.random.randint(0, len(df), size=size)
    return df.iloc[indices].reset_index(drop=True)


def sample_panel_with_replacement(
    df: pd.DataFrame, individual_map: pd.DataFrame, size: int | None = None
) -> pd.DataFrame:
    """
    Draws a sample of individuals with replacement from a panel dataset.

    :param df: The input DataFrame representing the full dataset.
    :param individual_map: A DataFrame mapping each individual ID to (start, end) row indices.
    :param size: The number of individuals to sample. Defaults to the number of individuals in the map.
    :return: A new DataFrame with the sampled individuals' rows, with reset index.
    :raises BiogemeError: if the individual_map is missing or empty.
    """
    if individual_map is None or individual_map.empty:
        raise BiogemeError("Panel individual map is missing or empty.")
    if size is None:
        size = len(individual_map)
    sampled_rows = []
    sampled_ids = np.random.choice(individual_map.index, size=size, replace=True)
    for individual_id in sampled_ids:
        start_idx, end_idx = individual_map.loc[individual_id]
        rows = df.loc[start_idx:end_idx]
        sampled_rows.append(rows)
    return pd.concat(sampled_rows, ignore_index=True)


def split_validation_sets(
    df: pd.DataFrame, slices: int, group_column: str | None = None
) -> list[tuple[pd.DataFrame, pd.DataFrame]]:
    """
    Splits a DataFrame into multiple (estimation, validation) pairs for cross-validation.

    :param df: The input DataFrame to split.
    :param slices: The number of folds (must be >= 2).
    :param group_column: Optional column name used to group rows (e.g., individual ID).
                         If provided, groups are kept together in folds.
    :return: A list of (estimation, validation) DataFrame tuples.
    :raises BiogemeError: if the number of slices is less than 2 or group column is not found.
    """
    if slices < 2:
        raise BiogemeError("Validation requires at least 2 slices.")
    if group_column is None:
        shuffled = df.sample(frac=1)
        folds = np.array_split(shuffled, slices)
    else:
        if group_column not in df.columns:
            raise BiogemeError(f"Grouping column '{group_column}' not found.")
        ids = df[group_column].unique()
        np.random.shuffle(ids)
        folds = [
            df[df[group_column].isin(group.tolist())]
            for group in np.array_split(ids, slices)
        ]
    estimation_sets = []
    validation_sets = []
    for i, validation in enumerate(folds):
        estimation = pd.concat(folds[:i] + folds[i + 1 :])
        estimation_sets.append(estimation.reset_index(drop=True))
        validation_sets.append(validation.reset_index(drop=True))
    return list(zip(estimation_sets, validation_sets))



"""Arithmetic expressions accepted by Biogeme: binary operators

Michel Bierlaire
Wed Mar 26 09:55:46 2025
"""

from __future__ import annotations

import logging

from .base_expressions import Expression, ExpressionOrNumeric
from .convert import validate_and_convert

logger = logging.getLogger(__name__)


class BinaryOperator(Expression):
    """
    Base class for arithmetic expressions that are binary operators.
    This expression is the result of the combination of two expressions,
    typically addition, subtraction, multiplication or division.
    """

    def __init__(self, left: ExpressionOrNumeric, right: ExpressionOrNumeric):
        """Constructor

        :param left: first arithmetic expression
        :type left: biogeme.expressions.Expression

        :param right: second arithmetic expression
        :type right: biogeme.expressions.Expression

        :raise BiogemeError: if one of the expressions is invalid, that is
            neither a numeric value nor a
            biogeme.expressions.Expression object.

        """
        super().__init__()
        self.left = validate_and_convert(left)
        self.right = validate_and_convert(right)

        self.children.append(self.left)
        self.children.append(self.right)



"""Simple tuple storing validity information

:author: Michel Bierlaire
:date: Fri Aug 18 16:36:22 2023

Stores the validity status and the explanation if invalid.
"""
from typing import NamedTuple


class Validity(NamedTuple):
    status: bool
    reason: str



"""Class that provides the function to the optimization algorithm

Michel Bierlaire
Sat Mar 29 16:27:47 2025
"""

import logging
from typing import Callable

import numpy as np
from biogeme_optimization.function import FunctionData, FunctionToMinimize

from biogeme.exceptions import BiogemeError
from biogeme.function_output import FunctionOutput

logger = logging.getLogger(__name__)


class NegativeLikelihood(FunctionToMinimize):
    """Provides the value of the function to be minimized, as well as its
    derivatives. To be used by the optimization package.

    """

    def __init__(
        self,
        dimension: int,
        loglikelihood: Callable[[np.ndarray, bool, bool, bool], FunctionOutput],
        parameters: dict[str, float] | None = None,
    ):
        """Constructor"""

        tolerance = None
        steptol = None
        if parameters is not None:
            if 'tolerance' in parameters:
                tolerance = parameters['tolerance']
            if 'steptol' in parameters:
                steptol = parameters['steptol']

        super().__init__(epsilon=tolerance, steptol=steptol)

        self.the_dimension: int = dimension  #: number of parameters to estimate

        self.loglikelihood: Callable[[np.ndarray, bool, bool, bool], FunctionOutput] = (
            loglikelihood
        )
        self.filename_for_best_iteration: str | None = None
        self.free_beta_names: list[str] | None = None
        self.best_value: float | None = None

    def save_iterations(
        self, filename_for_best_iteration: str, free_betas_names: list[str]
    ) -> None:
        self.filename_for_best_iteration = filename_for_best_iteration
        self.free_beta_names = free_betas_names

    def dimension(self) -> int:
        """Provides the number of variables of the problem"""
        return self.the_dimension

    def _f(self) -> float:
        if self.x is None:
            raise BiogemeError('The variables must be set first.')
        results = self.loglikelihood(self.x, gradient=False, hessian=False, bhhh=False)
        self._save_best_iteration(self.x, results.function)
        return -results.function

    def _f_g(self) -> FunctionData:
        if self.x is None:
            raise BiogemeError('The variables must be set first.')
        the_function_output: FunctionOutput = self.loglikelihood(
            self.x, gradient=True, hessian=False, bhhh=False
        )
        self._save_best_iteration(self.x, the_function_output.function)
        return FunctionData(
            function=-the_function_output.function,
            gradient=-the_function_output.gradient,
            hessian=None,
        )

    def _f_g_h(self) -> FunctionData:
        if self.x is None:
            raise BiogemeError('The variables must be set first.')

        the_function_output: FunctionOutput = self.loglikelihood(
            self.x, gradient=True, hessian=True, bhhh=False
        )
        self._save_best_iteration(self.x, the_function_output.function)
        return FunctionData(
            function=-the_function_output.function,
            gradient=-the_function_output.gradient,
            hessian=-the_function_output.hessian,
        )

    def _save_best_iteration(self, x: np.ndarray, f: float) -> None:
        if self.filename_for_best_iteration is None:
            return
        if self.best_value is None:
            self.best_value = f
        if f >= self.best_value:
            self.best_value = f
            with open(
                self.filename_for_best_iteration,
                "w",
                encoding="utf-8",
            ) as pf:
                for i, v in enumerate(x):
                    print(
                        f"{self.free_beta_names[i]} = {v}",
                        file=pf,
                    )



import os
import unittest
from datetime import timedelta

from biogeme.results_processing import (
    EstimationResults,
    RawEstimationResults,
    get_f12,
    generate_f12_file,
)


class TestF12Generation(unittest.TestCase):

    def setUp(self):
        # Setting up real estimation results using RawEstimationResults
        raw_estimation = RawEstimationResults(
            model_name='test_model',
            user_notes='user notes',
            beta_names=['beta1', 'beta2'],
            beta_values=[1.2, 1.3],
            lower_bounds=[-10, -11],
            upper_bounds=[None, 23],
            gradient=[0.001, -0.002],
            hessian=[[1, 2], [3, 4]],
            bhhh=[[10, 20], [30, 40]],
            null_log_likelihood=-1000,
            initial_log_likelihood=-2000,
            final_log_likelihood=-500,
            data_name='test_data',
            sample_size=1234,
            number_of_observations=987,
            monte_carlo=True,
            number_of_draws=10_000,
            types_of_draws={'var1': 'UNIFORM'},
            number_of_excluded_data=10,
            draws_processing_time=timedelta(seconds=10),
            optimization_messages={'Diagnostic': 'test', 'Number of iterations': 999},
            convergence=True,
            bootstrap=[[1.0, 1.0], [2.0, 2.0]],
            bootstrap_time=timedelta(seconds=20),
        )

        self.estimation_results = EstimationResults(
            raw_estimation_results=raw_estimation
        )

    def test_get_f12(self):
        # Generate the F12 output as a string
        f12_output = get_f12(self.estimation_results)
        # Check that the output contains relevant elements
        self.assertIn("test_model", f12_output)
        self.assertIn("From biogeme", f12_output)
        self.assertIn("END", f12_output)
        self.assertIn("beta1", f12_output)
        self.assertIn("beta2", f12_output)

        # Verify that parameter values are included
        self.assertIn("+1.200000000000e+00", f12_output)
        self.assertIn("+1.300000000000e+00", f12_output)

    def test_generate_f12_file(self):
        # Name of the file to generate
        filename = "test_f12_output.f12"

        # Remove file if it exists to ensure a clean test
        if os.path.exists(filename):
            os.remove(filename)

        # Generate the F12 file
        generate_f12_file(self.estimation_results, filename, overwrite=True)

        # Check that the file was created
        self.assertTrue(os.path.exists(filename))

        # Check the contents of the file
        with open(filename, 'r') as file:
            content = file.read()
            self.assertIn("test_model", content)
            self.assertIn("beta1", content)
            self.assertIn("+1.200000000000e+00", content)
            self.assertIn("999", content)

        # Clean up by removing the generated file
        os.remove(filename)


if __name__ == '__main__':
    unittest.main()



"""
Test the filenames module

:author: Michel Bierlaire
:date: Fri Feb 24 11:14:33 2023

"""
import os
import unittest
import biogeme.filenames as fn


class TestVersion(unittest.TestCase):
    def test_get_new_filename(self):
        the_name = fn.get_new_file_name('the_name', 'ext')
        self.assertEqual(the_name, 'the_name.ext')
        with open(the_name, 'w') as f:
            pass
        the_next_name = fn.get_new_file_name('the_name', 'ext')
        self.assertEqual(the_next_name, 'the_name~00.ext')
        os.remove(the_name)



"""

Mixture of logit
================

Example of the use of different algorithms to estimate the model.

:author: Michel Bierlaire, EPFL
:date: Sun Apr  9 17:38:34 2023

"""

import itertools
import pandas as pd
from biogeme.tools import format_timedelta
import biogeme.biogeme_logging as blog
import biogeme.biogeme as bio
from biogeme import models
import biogeme.exceptions as excep
from biogeme.expressions import Beta, bioDraws, log, MonteCarlo

# %%
# See the data processing script: :ref:`swissmetro_data`.
from swissmetro_data import (
    database,
    CHOICE,
    SM_AV,
    CAR_AV_SP,
    TRAIN_AV_SP,
    TRAIN_TT_SCALED,
    TRAIN_COST_SCALED,
    SM_TT_SCALED,
    SM_COST_SCALED,
    CAR_TT_SCALED,
    CAR_CO_SCALED,
)

logger = blog.get_screen_logger(level=blog.INFO)
logger.info('Example b05normal_mixture_all_algos.py')

# %%
# Parameters to be estimated
ASC_CAR = Beta('ASC_CAR', 0, None, None, 0)
ASC_TRAIN = Beta('ASC_TRAIN', 0, None, None, 0)
ASC_SM = Beta('ASC_SM', 0, None, None, 1)
B_COST = Beta('B_COST', 0, None, None, 0)

# %%
# Define a random parameter, normally distributed, designed to be used
# for Monte-Carlo simulation.
B_TIME = Beta('B_TIME', 0, None, None, 0)

# %%
# It is advised not to use 0 as starting value for the following parameter.
B_TIME_S = Beta('B_TIME_S', 1, None, None, 0)
B_TIME_RND = B_TIME + B_TIME_S * bioDraws('B_TIME_RND', 'NORMAL')

# %%
# Definition of the utility functions.
V1 = ASC_TRAIN + B_TIME_RND * TRAIN_TT_SCALED + B_COST * TRAIN_COST_SCALED
V2 = ASC_SM + B_TIME_RND * SM_TT_SCALED + B_COST * SM_COST_SCALED
V3 = ASC_CAR + B_TIME_RND * CAR_TT_SCALED + B_COST * CAR_CO_SCALED

# %%
# Associate utility functions with the numbering of alternatives
V = {1: V1, 2: V2, 3: V3}

# %%
# Associate the availability conditions with the alternatives
av = {1: TRAIN_AV_SP, 2: SM_AV, 3: CAR_AV_SP}

# %%
# Conditional to B_TIME_RND, we have a logit model (called the kernel)
prob = models.logit(V, av, CHOICE)

# %%
# We integrate over B_TIME_RND using Monte-Carlo
logprob = log(MonteCarlo(prob))

# %%
# Options for the optimization algorithm
# --------------------------------------

# %%
# The conjugate gradient iteration can be constrained to stay feasible, or not.
infeasible_cg_values = [True, False]

# %%
# The radius of the first trust region is tested with three different values.
initial_radius_values = [0.1, 1.0, 10.0]

# %%
# The percentage of iterations such that the analytical second derivatives is evaluated.
second_derivatives_values = [0.0, 0.5, 1.0]

# %%
# We run the optimization algorithm with all possible combinations of the parameters. The results are stored in a Pandas DataFrame called ``summary``.
results = {}
summary = pd.DataFrame(
    columns=[
        'LogLikelihood',
        'GradientNorm',
        'Optimization time',
        'TerminationCause',
        'Status',
    ]
)

for infeasible_cg, initial_radius, second_derivatives in itertools.product(
    infeasible_cg_values, initial_radius_values, second_derivatives_values
):
    # Create the Biogeme object
    the_biogeme = bio.BIOGEME(database, logprob, parameter_file='few_draws.toml')
    # We set the parameters of the optimization algorithm
    the_biogeme.infeasible_cg = infeasible_cg
    the_biogeme.initial_radius = initial_radius
    the_biogeme.second_derivatives = second_derivatives
    # We cancel the generation of the outputfiles
    the_biogeme.generate_html = False
    the_biogeme.generate_pickle = False

    name = (
        f'cg_{infeasible_cg}_radius_{initial_radius}_second_deriv_{second_derivatives}'
    )
    the_biogeme.modelName = f'b05normal_mixture_algo_{name}'.strip()
    result_data = {
        'InfeasibleCG': infeasible_cg,
        'InitialRadius': initial_radius,
        'SecondDerivatives': second_derivatives,
        'Status': 'Success',  # Assume success unless an exception is caught
    }

    try:
        results[name] = the_biogeme.estimate()
        opt_time = format_timedelta(
            results[name].data.optimizationMessages["Optimization time"]
        )

        result_data.update(
            {
                'LogLikelihood': results[name].data.logLike,
                'GradientNorm': results[name].data.gradientNorm,
                'Optimization time': opt_time,
                'TerminationCause': results[name].data.optimizationMessages[
                    "Cause of termination"
                ],
            }
        )

    except excep.BiogemeError as e:
        print(e)
        result_data.update(
            {
                'Status': 'Failed',
                'LogLikelihood': None,
                'GradientNorm': None,
                'Optimization time': None,
                'TerminationCause': str(e),
            }
        )
        results[name] = None

    summary = pd.concat([summary, pd.DataFrame([result_data])], ignore_index=True)

# %%
summary

# %%
SUMMARY_FILE = '05normalMixture_allAlgos.csv'
summary.to_csv(SUMMARY_FILE, index=False)
print(f'Summary reported in file {SUMMARY_FILE}')



"""
Tools to transform NaN in numpy arrays in order to be serialized

Michel Bierlaire
Mon May 19 2025, 11:57:38
"""

import numpy as np


def safe_serialize_array(
    array: np.ndarray,
) -> list[float | None] | list[list[float | None]]:
    """
    Convert a NumPy array with potential NaN values into a nested or flat list
    with `None` in place of `np.nan`, making it safe for YAML or JSON serialization.

    :param array: A NumPy array that may contain np.nan values.
    :return: A list (1D or 2D) with None in place of np.nan, suitable for serialization.
    """
    if not isinstance(array, np.ndarray):
        raise TypeError(f'Input must be a NumPy array, not {type(array)}.')
    if array.ndim == 1:
        return [None if np.isnan(val) else float(val) for val in array]
    elif array.ndim == 2:
        return [[None if np.isnan(val) else float(val) for val in row] for row in array]
    else:
        raise ValueError("Only 1D and 2D arrays are supported.")


def safe_deserialize_array(
    serialized: list[float | None] | list[list[float | None]],
) -> list[float] | list[list[float]]:
    """
    Convert a flat or nested list with None values (as parsed from YAML or JSON)
    into a list with None replaced by float('nan').

    :param serialized: A list (1D or 2D) containing float or None values.
    :return: A list (1D or 2D) with None replaced by float('nan').
    """
    if not isinstance(serialized, list):
        raise TypeError("Input must be a list.")

    if all(isinstance(val, (float, int, type(None))) for val in serialized):
        # 1D case
        return [float('nan') if val is None else float(val) for val in serialized]
    elif all(
        isinstance(row, list)
        and all(isinstance(val, (float, int, type(None))) for val in row)
        for row in serialized
    ):
        # 2D case
        return [
            [float('nan') if val is None else float(val) for val in row]
            for row in serialized
        ]
    else:
        raise TypeError("Input must be a 1D or 2D list of numbers or None.")



"""

Monte-Carlo integration
=======================

Calculation of a mixtures of logit models where the integral is
calculated using numerical integration and Monte-Carlo integration
with various types of draws.

:author: Michel Bierlaire, EPFL
:date: Thu Apr 13 20:58:50 2023
"""


import biogeme.biogeme as bio
from biogeme import models
import biogeme.distributions as dist
from biogeme.expressions import (
    Expression,
    Integrate,
    RandomVariable,
    MonteCarlo,
    bioDraws,
)

from swissmetro_one import (
    database,
    TRAIN_TT_SCALED,
    TRAIN_COST_SCALED,
    SM_TT_SCALED,
    SM_COST_SCALED,
    CAR_TT_SCALED,
    CAR_CO_SCALED,
    TRAIN_AV_SP,
    SM_AV,
    CAR_AV_SP,
    CHOICE,
)

# %%
R = 10000

# %%
# Parameters
ASC_CAR = 0.137
ASC_TRAIN = -0.402
ASC_SM = 0
B_TIME = -2.26
B_TIME_S = 1.66
B_COST = -1.29

# %%
# Generate several versions of the error component.
omega = RandomVariable('omega')
density = dist.normalpdf(omega)
B_TIME_RND = B_TIME + B_TIME_S * omega
B_TIME_RND_normal = B_TIME + B_TIME_S * bioDraws('B_NORMAL', 'NORMAL')
B_TIME_RND_anti = B_TIME + B_TIME_S * bioDraws('B_ANTI', 'NORMAL_ANTI')
B_TIME_RND_halton = B_TIME + B_TIME_S * bioDraws('B_HALTON', 'NORMAL_HALTON2')
B_TIME_RND_mlhs = B_TIME + B_TIME_S * bioDraws('B_MLHS', 'NORMAL_MLHS')
B_TIME_RND_antimlhs = B_TIME + B_TIME_S * bioDraws('B_ANTIMLHS', 'NORMAL_MLHS_ANTI')


# %%
def logit(the_b_time_rnd: Expression) -> Expression:
    """
    Calculate the conditional logit model for a given random parameter.

    :param the_b_time_rnd: expression for the random parameter.

    :return: logit model to be integrated.
    """
    v_1 = ASC_TRAIN + the_b_time_rnd * TRAIN_TT_SCALED + B_COST * TRAIN_COST_SCALED
    v_2 = ASC_SM + the_b_time_rnd * SM_TT_SCALED + B_COST * SM_COST_SCALED
    v_3 = ASC_CAR + the_b_time_rnd * CAR_TT_SCALED + B_COST * CAR_CO_SCALED

    # Associate utility functions with the numbering of alternatives
    v = {1: v_1, 2: v_2, 3: v_3}

    # Associate the availability conditions with the alternatives
    av = {1: TRAIN_AV_SP, 2: SM_AV, 3: CAR_AV_SP}

    # The choice model is a logit, with availability conditions
    integrand = models.logit(v, av, CHOICE)
    return integrand


# %%
# Generate each integral.
numericalI = Integrate(logit(B_TIME_RND) * density, 'omega')
normal = MonteCarlo(logit(B_TIME_RND_normal))
anti = MonteCarlo(logit(B_TIME_RND_anti))
halton = MonteCarlo(logit(B_TIME_RND_halton))
mlhs = MonteCarlo(logit(B_TIME_RND_mlhs))
antimlhs = MonteCarlo(logit(B_TIME_RND_antimlhs))

# %%
simulate = {
    'Numerical': numericalI,
    'MonteCarlo': normal,
    'Antithetic': anti,
    'Halton': halton,
    'MLHS': mlhs,
    'Antithetic MLHS': antimlhs,
}

# %%
biosim = bio.BIOGEME(database, simulate)
biosim.number_of_draws = R

# %%
results = biosim.simulate(theBetaValues={})
results

# %%
print(f'Number of draws: {R}')
for c in results.columns:
    print(f'{c}:\t{results.loc[0,c]}')



import unittest

from biogeme.catalog import Controller
from biogeme.exceptions import BiogemeError


class TestController(unittest.TestCase):
    def setUp(self):
        self.controller_name = 'my_controller'
        self.specification_names = ('spec_1', 'spec_2')
        self.the_controller = Controller(
            controller_name=self.controller_name,
            specification_names=self.specification_names,
        )

    def test_constructor(self):
        self.assertEqual(self.the_controller.controller_name, self.controller_name)
        self.assertTupleEqual(
            self.the_controller.specification_names, self.specification_names
        )
        self.assertEqual(self.the_controller.current_index, 0)
        self.assertTupleEqual(
            tuple(self.the_controller.dict_of_index.keys()), self.specification_names
        )
        self.assertListEqual(self.the_controller.controlled_catalogs, [])
        self.assertEqual(self.the_controller.controller_size(), 2)

    def test_all_configurations(self):
        all_configurations = self.the_controller.all_configurations()
        expected_result = {'my_controller:spec_1', 'my_controller:spec_2'}
        self.assertSetEqual(all_configurations, expected_result)

    def test_name(self):
        name = 'spec_2'
        self.the_controller.set_name(name)
        current_name = self.the_controller.current_name()
        self.assertEqual(name, current_name)

        incorrect_name = 'incorrect_name'
        with self.assertRaises(BiogemeError):
            self.the_controller.set_name(incorrect_name)

    def test_index(self):
        self.the_controller.set_index(1)
        current_name = self.the_controller.current_name()
        expected_name = 'spec_2'
        self.assertEqual(expected_name, current_name)
        self.the_controller.reset_selection()
        current_name = self.the_controller.current_name()
        expected_name = 'spec_1'
        self.assertEqual(expected_name, current_name)

    def test_modify_controller(self):
        self.the_controller.reset_selection()
        self.the_controller.modify_controller(step=1, circular=False)
        self.assertEqual(self.the_controller.current_index, 1)
        self.the_controller.modify_controller(step=1, circular=False)
        self.assertEqual(self.the_controller.current_index, 1)
        self.the_controller.modify_controller(step=1, circular=True)
        self.assertEqual(self.the_controller.current_index, 0)
        self.the_controller.modify_controller(step=10, circular=False)
        self.assertEqual(self.the_controller.current_index, 1)
        self.the_controller.modify_controller(step=-10, circular=False)
        self.assertEqual(self.the_controller.current_index, 0)


if __name__ == '__main__':
    unittest.main()



"""

Heteroscedastic specification
=============================

 Illustrates a heteroscedastic specification. A different scale is
 associated with different segments of the sample.

:author: Michel Bierlaire, EPFL
:date: Sun Apr  9 17:23:03 2023


"""

import biogeme.biogeme as bio
from biogeme import models
from biogeme.expressions import Beta

# %%
# See the data processing script: :ref:`swissmetro_data`.
from swissmetro_data import (
    database,
    CHOICE,
    SM_AV,
    CAR_AV_SP,
    TRAIN_AV_SP,
    TRAIN_TT_SCALED,
    TRAIN_COST_SCALED,
    SM_TT_SCALED,
    SM_COST_SCALED,
    CAR_TT_SCALED,
    CAR_CO_SCALED,
    GROUP,
)


# %%
# Parameters to be estimated.
ASC_CAR = Beta('ASC_CAR', 0, None, None, 0)
ASC_TRAIN = Beta('ASC_TRAIN', 0, None, None, 0)
ASC_SM = Beta('ASC_SM', 0, None, None, 1)
B_TIME = Beta('B_TIME', 0, None, None, 0)
B_COST = Beta('B_COST', 0, None, None, 0)
Scale_group3 = Beta('Scale_group3', 1, 0.001, None, 0)

# %%
# Definition of the utility functions.
V1 = ASC_TRAIN + B_TIME * TRAIN_TT_SCALED + B_COST * TRAIN_COST_SCALED
V2 = ASC_SM + B_TIME * SM_TT_SCALED + B_COST * SM_COST_SCALED
V3 = ASC_CAR + B_TIME * CAR_TT_SCALED + B_COST * CAR_CO_SCALED

# %%
# Scale associated with group 3 is estimated.
scale = (GROUP != 3) + (GROUP == 3) * Scale_group3

# %%
# Scale the utility functions, and associate them with the numbering
# of alternatives.
V = {1: scale * V1, 2: scale * V2, 3: scale * V3}

# %%
# Associate the availability conditions with the alternatives.
av = {1: TRAIN_AV_SP, 2: SM_AV, 3: CAR_AV_SP}

# %%
# Definition of the model. This is the contribution of each
# observation to the log likelihood function.
logprob = models.loglogit(V, av, CHOICE)

# %%
# These notes will be included as such in the report file.
USER_NOTES = (
    'Illustrates heteroscedastic specification. A different scale is'
    ' associated with different segments of the sample.'
)

# %%
# Create the Biogeme object.
the_biogeme = bio.BIOGEME(database, logprob, user_notes=USER_NOTES)
the_biogeme.modelName = 'b03scale'

# %%
# Estimate the parameters.
results = the_biogeme.estimate()

# %%
print(results.short_summary())

# %%
# Get the results in a pandas table
pandas_results = results.getEstimatedParameters()
pandas_results



"""
Stub file for PyCharm and static analyzers.

This file declares attributes of the BIOGEME class that are dynamically
added at runtime (e.g., through setattr or property injection), so that
IDEs like PyCharm can provide autocompletion and suppress false warnings.
"""

class BIOGEME:
    seed: int  # Dynamically injected parameter (e.g., via setattr or property)
    tolerance: float  # Dynamically injected parameter (e.g., via setattr or property)
    steptol: float  # Dynamically injected parameter (e.g., via setattr or property)
    max_iterations: (
        int  # Dynamically injected parameter (e.g., via setattr or property)
    )
    number_of_threads: (
        int  # Dynamically injected parameter (e.g., via setattr or property)
    )
    enlarging_factor: (
        float  # Dynamically injected parameter (e.g., via setattr or property)
    )
    initial_radius: (
        float  # Dynamically injected parameter (e.g., via setattr or property)
    )
    infeasible_cg: (
        bool  # Dynamically injected parameter (e.g., via setattr or property)
    )
    second_derivatives: (
        int  # Dynamically injected parameter (e.g., via setattr or property)
    )
    dogleg: bool  # Dynamically injected parameter (e.g., via setattr or property)
    bootstrap_samples: (
        int  # Dynamically injected parameter (e.g., via setattr or property)
    )
    save_iterations: (
        bool  # Dynamically injected parameter (e.g., via setattr or property)
    )
    generate_html: (
        bool  # Dynamically injected parameter (e.g., via setattr or property)
    )
    generate_yaml: (
        bool  # Dynamically injected parameter (e.g., via setattr or property)
    )
    optimization_algorithm: (
        str  # Dynamically injected parameter (e.g., via setattr or property)
    )
    maximum_number_catalog_expressions: (
        int  # Dynamically injected parameter (e.g., via setattr or property)
    )



"""Defines a class that characterized the context to apply sampling of alternatives

:author: Michel Bierlaire
:date: Wed Sep  6 14:38:31 2023
"""

import logging
from collections.abc import Iterable
from dataclasses import dataclass
from typing import NamedTuple

import pandas as pd

from biogeme.exceptions import BiogemeError
from biogeme.expressions import (
    Expression,
    list_of_variables_in_expression,
)
from biogeme.nests import NestsForCrossNestedLogit
from biogeme.partition import Partition, Segment

logger = logging.getLogger(__name__)

MEV_PREFIX = '_MEV_'
LOG_PROBA_COL = '_log_proba'
MEV_WEIGHT = '_mev_weight'
CNL_PREFIX = '_CNL_'


class StratumTuple(NamedTuple):
    """A stratum is an element of a partition of the full choice set,
    combined with the number of alternatives that must be sampled.
    """

    subset: Segment
    sample_size: int


class CrossVariableTuple(NamedTuple):
    """A cross variable is a variable that involves socio-economic
    attributes of the individuals, and attributes of the
    alternatives. It can only be calculated after the sampling has
    been made.
    """

    name: str
    formula: Expression


@dataclass
class SamplingContext:
    """Class gathering the data needed to perform an estimation with
    samples of alternatives

    :param the_partition: Partition used for the sampling.

    :param sample_sizes: number of alternative to draw from each segment.

    :param individuals: Pandas data frame containing all the
        individuals as rows. One column must contain the choice of
        each individual.

    :param choice_column: name of the column containing the choice of
        each individual.

    :param alternatives: Pandas data frame containing all the
        alternatives as rows. One column must contain a unique ID
        identifying the alternatives. The other columns contain
        variables to include in the data file.

    :param id_column: name of the column containing the Ids of the alternatives.

    :param utility_function: definition of the generic utility function

    :param combined_variables: definition of interaction variables

    :param mev_partition: If a second choice set need to be sampled
        for the MEV terms, the corresponding partition is provided
        here.

    """

    the_partition: Partition
    sample_sizes: Iterable[int]
    individuals: pd.DataFrame
    choice_column: str
    alternatives: pd.DataFrame
    id_column: str
    biogeme_file_name: str
    utility_function: Expression
    combined_variables: list[CrossVariableTuple]
    mev_partition: Partition | None = None
    mev_sample_sizes: Iterable[int] | None = None
    cnl_nests: NestsForCrossNestedLogit | None = None

    def check_expression(self, expression: Expression) -> None:
        """Verifies if the variables contained in the expression can be found in the databases"""
        variables = {
            var.name
            for var in list_of_variables_in_expression(the_expression=expression)
        }
        for variable in variables:
            if (
                variable not in self.individuals.columns
                and variable not in self.alternatives.columns
                and all(variable != t.name for t in self.combined_variables)
            ):
                error_msg = (
                    f'Invalid expression. Variable "{variable}" has not been found in '
                    f'the provided database'
                )
                raise BiogemeError(error_msg)

    def check_partition(self) -> None:
        """Check if the partition is truly a partition. If not, an exception is raised

        :raise BiogemeError: if some elements are present in more than one subset.

        :raise BiogemeError: if the size of the union of the subsets does
           not match the expected total size

        :raise BiogemeError: if an alternative in the partition does
            not appear in the database of alternatives

        :raise BiogemeError: if a segment is empty

        :raise BiogemeError: if the number of sampled alternatives in
            a stratum is incorrect , that is zero, or larger than the
            stratum size..

        """
        # Verify that all requested alternatives appear in the database of alternatives
        for stratum in self.sampling_protocol:
            n = len(stratum.subset)
            if n == 0:
                error_msg = 'A stratum is empty'
                raise BiogemeError(error_msg)
            k = stratum.sample_size
            if k > n:
                error_msg = f'Cannot draw {k} elements in a stratum of size {n}'
                raise BiogemeError(error_msg)

            if k == 0:
                error_msg = 'At least one alternative must be selected in each segment'
                raise BiogemeError(error_msg)

            for alt in stratum.subset:
                if alt not in self.alternatives[self.id_column].values:
                    error_msg = (
                        f'Alternative {alt} does not appear in the database of '
                        f'alternatives'
                    )
                    raise BiogemeError(error_msg)

    def check_mev_partition(self) -> None:
        """Check if the partition is a partition of the MEV
        alternatives. It does not need to cover the full choice set"""

        if self.mev_partition:
            if self.mev_sample_sizes is None:
                error_msg = (
                    'If mev_partition is defined, mev_sample_size must also be defined'
                )
                raise BiogemeError(error_msg)

        if self.mev_sample_sizes:
            if self.mev_partition is None:
                error_msg = (
                    'If mev_sample_sizes is defined, mev_partition must also be defined'
                )
                raise BiogemeError(error_msg)

        if self.cnl_nests and self.mev_partition:
            if self.cnl_nests.mev_alternatives != self.mev_partition.full_set:
                in_nest_not_in_partition = (
                    self.cnl_nests.mev_alternatives - self.mev_partition.full_set
                )
                in_partition_not_in_nest = (
                    self.mev_partition.full_set - self.cnl_nests.mev_alternatives
                )
                error_msg = ''
                if in_nest_not_in_partition:
                    error_msg += (
                        f'The following alternative(s) belong to a nest but not to the'
                        f' partition for the sample: {in_nest_not_in_partition}. '
                    )
                if in_partition_not_in_nest:
                    error_msg += (
                        f'The following alternative(s) belong to the partition for '
                        f'the MEV sample, but not to any nest: {in_partition_not_in_nest}'
                    )

    def check_valid_alternatives(self, set_of_ids: set[int]) -> None:
        """Check if the IDs in set are indeed valid
            alternatives. Typically used to check if a nest is well
            defined

        :param set_of_ids: set of identifiers to check

        :raise BiogemeError: if at least one id is invalid.
        """
        if (
            not pd.Series(list(set_of_ids))
            .isin(self.alternatives[self.id_column])
            .all()
        ):
            missing_values = set_of_ids - set(self.alternatives[self.id_column])
            raise BiogemeError(
                f'The following IDs are not valid alternative IDs: {missing_values}'
            )

    def include_cnl_alphas(self) -> None:
        if self.cnl_nests is None:
            return
        for nest in self.cnl_nests:
            column_name = f'{CNL_PREFIX}{nest.name}'
            self.alternatives[column_name] = self.alternatives[self.id_column].map(
                lambda x: (
                    self.cnl_nests.get_alpha_values(alternative_id=x)[nest.name]
                    if x in nest.dict_of_alpha
                    else 0.0
                )
            )

    def __post_init__(self) -> None:
        # Check for empty utility function
        if self.utility_function is None:
            raise BiogemeError('No utility function has been provided')

        # Check for empty strings
        if not self.choice_column:
            raise BiogemeError('choice_column should not be an empty string.')

        if not self.id_column:
            raise BiogemeError('id_column should not be an empty string.')

        # Validate that the DataFrames are not empty
        if self.individuals.empty or self.alternatives.empty:
            raise BiogemeError(
                'DataFrames individuals or alternatives should not be empty.'
            )

        self.check_partition()
        logger.debug('Check if there is a MEV partition')
        if self.mev_partition or self.mev_sample_sizes:
            logger.debug('Yes, there is a MEV partition')
            self.check_mev_partition()
        else:
            logger.debug('No, there is no MEV partition')

        # If CNL nests are defined, check that the alphas are all
        # fixed and that the nests have a name.
        if self.cnl_nests:
            if not self.cnl_nests.all_alphas_fixed():
                error_msg = 'For the CNL model, all alpha parameters must be fixed.'
                raise BiogemeError(error_msg)
            if not self.cnl_nests.check_names():
                error_msg = 'For the CNL model, all nests must have a name.'
                raise BiogemeError(error_msg)
        self.number_of_alternatives = self.alternatives.shape[0]
        self.number_of_individuals = self.individuals.shape[0]

        # Validate that choice_column is in the individuals DataFrame
        if self.choice_column not in self.individuals.columns:
            raise BiogemeError(
                f'{self.choice_column} is not a column in the individuals DataFrame.'
            )

        # Validate that id_column is in the alternatives DataFrame
        if self.id_column not in self.alternatives.columns:
            raise BiogemeError(
                f'{self.id_column} is not a column in the alternatives DataFrame.'
            )

        # Check for data types
        if not self.individuals[self.choice_column].dtype in [int, float]:
            raise BiogemeError(
                f'Column {self.choice_column} in data frame "individuals" should '
                f'be of type int or float.'
            )

        if not self.alternatives[self.id_column].dtype in [int, float]:
            raise BiogemeError(
                f'Column {self.id_column} in alternatives should be of type int or float.'
            )

        self.check_expression(self.utility_function)
        for cross_variable in self.combined_variables:
            self.check_expression(cross_variable.formula)

        self.include_cnl_alphas()

    @property
    def attributes(self) -> set[str]:
        """List of attributes for the choice model"""
        return set(self.alternatives.columns) | {
            combined_variable.name for combined_variable in self.combined_variables
        }

    @property
    def mev_prefix(self) -> str:
        """Build the prefix for the MEV columns"""
        return '' if self.mev_partition is None else MEV_PREFIX

    @property
    def sampling_protocol(self) -> list[StratumTuple]:
        """Provides a list of strata characterizing the sampling"""
        return [
            StratumTuple(subset=segment, sample_size=size)
            for segment, size in zip(self.the_partition, self.sample_sizes)
        ]

    @property
    def mev_sampling_protocol(self) -> list[StratumTuple] | None:
        """Provides a list of strata characterizing the MEV sampling"""
        if self.mev_partition is None:
            return None
        return [
            StratumTuple(subset=segment, sample_size=size)
            for segment, size in zip(self.mev_partition, self.mev_sample_sizes)
        ]

    @property
    def total_sample_size(self) -> int:
        """Sample size"""
        return sum(stratum.sample_size for stratum in self.sampling_protocol)

    @property
    def total_mev_sample_size(self) -> int:
        """Sample size"""
        if self.mev_partition is None:
            return 0
        return sum(stratum.sample_size for stratum in self.mev_sampling_protocol)

    def reporting(self) -> str:
        """Summarizes the configuration specified by the context object."""
        result = {
            'Size of the choice set': self.alternatives.shape[0],
            'Main partition': (
                f'{self.the_partition.number_of_segments()} segment(s) of size '
                f'{", ".join([str(len(segment)) for segment in self.the_partition])}'
            ),
            'Main sample': f'{self.total_sample_size}: ',
        }

        result['Main sample'] += ', '.join(
            [
                f'{stratum.sample_size}/{len(stratum.subset)}'
                for stratum in self.sampling_protocol
            ]
        )
        if self.mev_partition:
            result['Nbr of MEV alternatives'] = len(self.mev_partition.full_set)
            result['MEV partition'] = (
                f'{self.mev_partition.number_of_segments()} segment(s) of size '
                f'{", ".join([str(len(segment)) for segment in self.mev_partition])}'
            )
            result['MEV sample'] = f'{self.total_mev_sample_size}: '
            result['MEV sample'] += ', '.join(
                [
                    f'{stratum.sample_size}/{len(stratum.subset)}'
                    for stratum in self.mev_sampling_protocol
                ]
            )

        output = ''
        for section, description in result.items():
            output += f'{section}: {description}\n'
        return output



import unittest

from biogeme.catalog import Catalog, segmentation_catalogs
from biogeme.exceptions import BiogemeError
from biogeme.expressions import Beta, SELECTION_SEPARATOR, SEPARATOR
from biogeme.segmentation import DiscreteSegmentationTuple


class TestSegmentationCatalogs(unittest.TestCase):

    def setUp(self):
        self.beta = Beta('test_param', 1.0, None, None, 0)
        self.beta_list = [self.beta]
        self.valid_segmentation = DiscreteSegmentationTuple(
            variable='group', mapping={1: 'A', 2: 'B'}
        )

    def test_valid_catalogs_generation(self):
        catalogs = segmentation_catalogs(
            generic_name='test',
            beta_parameters=self.beta_list,
            potential_segmentations=(self.valid_segmentation,),
            maximum_number=1,
        )
        self.assertIsInstance(catalogs, list)
        self.assertTrue(all(isinstance(cat, Catalog) for cat in catalogs))
        self.assertEqual(len(catalogs), 1)
        self.assertTrue(catalogs[0].name.startswith('segmented_'))

    def test_invalid_beta_parameters_type(self):
        with self.assertRaisesRegex(
            BiogemeError, 'A list is expected for beta_parameters'
        ):
            segmentation_catalogs(
                generic_name='test',
                beta_parameters='not_a_list',
                potential_segmentations=(self.valid_segmentation,),
                maximum_number=1,
            )

    def test_segment_name_with_separator_raises(self):
        bad_segmentation = DiscreteSegmentationTuple(
            variable='group', mapping={1: f'bad{SEPARATOR}name'}
        )
        with self.assertRaisesRegex(BiogemeError, 'Invalid segment name for variable'):
            segmentation_catalogs(
                generic_name='test',
                beta_parameters=self.beta_list,
                potential_segmentations=(bad_segmentation,),
                maximum_number=1,
            )

    def test_segment_name_with_selection_separator_raises(self):
        bad_segmentation = DiscreteSegmentationTuple(
            variable='group', mapping={1: f'bad{SELECTION_SEPARATOR}name'}
        )
        with self.assertRaisesRegex(BiogemeError, 'Invalid segment name for variable'):
            segmentation_catalogs(
                generic_name='test',
                beta_parameters=self.beta_list,
                potential_segmentations=(bad_segmentation,),
                maximum_number=1,
            )

    def test_no_segmentation_case_included(self):
        catalogs = segmentation_catalogs(
            generic_name='test',
            beta_parameters=self.beta_list,
            potential_segmentations=(self.valid_segmentation,),
            maximum_number=0,
        )
        names = [named.name for named in catalogs[0].named_expressions]
        self.assertIn('no_seg', names)


if __name__ == '__main__':
    unittest.main()



"""

Read of estimate
================

Function to estimate the parameters, or read them from a file, if available.

:author: Michel Bierlaire, EPFL
:date: Sat Dec  9 17:20:49 2023
"""
import logging
from typing import Optional
from biogeme.biogeme import BIOGEME
from biogeme.results import bioResults
import biogeme.exceptions as excep

logger = logging.getLogger(__name__)


def read_or_estimate(
    the_biogeme: BIOGEME, directory: Optional[str] = '.'
) -> bioResults:
    """
    Function to estimate the parameters, or read them from a file, if available.

    :param the_biogeme: Biogeme object.
    :param directory: directory where the pickle file is supposed to be.

    :return: estimation results.
    """
    try:
        filename = f'{directory}/{the_biogeme.modelName}.pickle'
        logger.info('Results are read from the file {filename}.')
        results = bioResults(pickleFile=filename)
    except excep.FileNotFoundError:
        logger.info('Parameters are estimated.')
        results = the_biogeme.estimate()

    return results



from dataclasses import dataclass
import unittest
import pandas as pd

from biogeme.sampling_of_alternatives import (
    StratumTuple,
    CrossVariableTuple,
    SamplingContext,
)
from biogeme.partition import Partition
from biogeme.expressions import Beta, Variable
from biogeme.exceptions import BiogemeError


class TestSamplingContext(unittest.TestCase):
    def setUp(self):
        self.valid_individuals = pd.DataFrame(
            {'choice': [1, 2, 3], 'age': [22, 33, 44]}
        )

        self.valid_alternatives = pd.DataFrame(
            {
                'id': [1, 2, 3, 5, 6],
                'travel_time': [12, 23, 34, 45, 56],
                'travel_cost': [13, 11, 9, 7, 5],
            }
        )

        self.valid_partition = Partition([{1, 2}, {3, 5, 6}])
        self.valid_sample_sizes = [1, 2]

        self.valid_utility_function = Beta('beta_cost', 0, None, None, 0) * Variable(
            'travel_cost'
        ) + Beta('beta_age_time', 0, None, None, 0) * Variable('age_time')

        self.valid_combined_variables = [
            CrossVariableTuple('age_time', Variable('age') * Variable('travel_time'))
        ]

    def test_check_expression(self):
        context = SamplingContext(
            the_partition=self.valid_partition,
            sample_sizes=self.valid_sample_sizes,
            individuals=self.valid_individuals,
            choice_column='choice',
            alternatives=self.valid_alternatives,
            id_column='id',
            biogeme_file_name='test.biogeme',
            utility_function=self.valid_utility_function,
            combined_variables=self.valid_combined_variables,
        )

        expression = Variable('any_variable')
        with self.assertRaisesRegex(BiogemeError, 'Invalid expression'):
            context.check_expression(expression)

    def test_check_partition(self):
        # Test valid partition
        context = SamplingContext(
            the_partition=self.valid_partition,
            sample_sizes=self.valid_sample_sizes,
            individuals=self.valid_individuals,
            choice_column='choice',
            alternatives=self.valid_alternatives,
            id_column='id',
            biogeme_file_name='test.biogeme',
            utility_function=self.valid_utility_function,
            combined_variables=self.valid_combined_variables,
        )

        context.check_partition()

        # Test invalid partition (empty stratum)
        invalid_partition = [(), ('b', 'c')]
        with self.assertRaisesRegex(Exception, 'A stratum is empty'):
            context = SamplingContext(
                the_partition=invalid_partition,
                sample_sizes=self.valid_sample_sizes,
                individuals=self.valid_individuals,
                choice_column='choice',
                alternatives=self.valid_alternatives,
                id_column='id',
                biogeme_file_name='test.biogeme',
                utility_function=self.valid_utility_function,
                combined_variables=self.valid_combined_variables,
            )

        # Test invalid partition (alternative not in DB)
        invalid_partition = [('x',), ('b', 'c')]
        with self.assertRaisesRegex(
            Exception, 'Alternative x does not appear in the database of alternatives'
        ):
            context = SamplingContext(
                the_partition=invalid_partition,
                sample_sizes=self.valid_sample_sizes,
                individuals=self.valid_individuals,
                choice_column='choice',
                alternatives=self.valid_alternatives,
                id_column='id',
                biogeme_file_name='test.biogeme',
                utility_function=self.valid_utility_function,
                combined_variables=self.valid_combined_variables,
            )

    def test_check_mev_partition(self):
        # Test valid partition
        context = SamplingContext(
            the_partition=self.valid_partition,
            sample_sizes=self.valid_sample_sizes,
            individuals=self.valid_individuals,
            choice_column='choice',
            alternatives=self.valid_alternatives,
            id_column='id',
            biogeme_file_name='test.biogeme',
            utility_function=self.valid_utility_function,
            combined_variables=self.valid_combined_variables,
            mev_partition=self.valid_partition,
            mev_sample_sizes=self.valid_sample_sizes
        )

        context.check_mev_partition()

        # Test invalid partition
        with self.assertRaisesRegex(Exception, 'If mev_partition'):
            context = SamplingContext(
                the_partition=self.valid_partition,
                sample_sizes=self.valid_sample_sizes,
                individuals=self.valid_individuals,
                choice_column='choice',
                alternatives=self.valid_alternatives,
                id_column='id',
                biogeme_file_name='test.biogeme',
                utility_function=self.valid_utility_function,
                combined_variables=self.valid_combined_variables,
                mev_partition=self.valid_partition,
                mev_sample_sizes=None
            )

        with self.assertRaisesRegex(Exception, 'If mev_sample_sizes'):
            context = SamplingContext(
                the_partition=self.valid_partition,
                sample_sizes=self.valid_sample_sizes,
                individuals=self.valid_individuals,
                choice_column='choice',
                alternatives=self.valid_alternatives,
                id_column='id',
                biogeme_file_name='test.biogeme',
                utility_function=self.valid_utility_function,
                combined_variables=self.valid_combined_variables,
                mev_partition=None,
                mev_sample_sizes=self.valid_sample_sizes
            )
            
        # Test invalid partition (alternative not in DB)
        invalid_partition = [('x',), ('b', 'c')]
        with self.assertRaisesRegex(Exception, 'Alternative x does not appear in the database of alternatives'):
            context = SamplingContext(
                the_partition=invalid_partition,
                sample_sizes=self.valid_sample_sizes,
                individuals=self.valid_individuals,
                choice_column='choice',
                alternatives=self.valid_alternatives,
                id_column='id',
                biogeme_file_name='test.biogeme',
                utility_function=self.valid_utility_function,
                combined_variables=self.valid_combined_variables,
                mev_partition=invalid_partition,
                mev_sample_sizes=self.valid_sample_sizes
            )


if __name__ == '__main__':
    unittest.main()



import unittest
import warnings

import numpy as np

from biogeme.multiobjectives import (
    AIC_BIC_dimension,
    aic_bic_dimension,
    loglikelihood_dimension,
)


class DummyResults:
    def __init__(
        self,
        final_log_likelihood=None,
        number_of_parameters=None,
        akaike_information_criterion=None,
        bayesian_information_criterion=None,
        raw_estimation_results=True,
    ):
        self.final_log_likelihood = final_log_likelihood
        self.number_of_parameters = number_of_parameters
        self.akaike_information_criterion = akaike_information_criterion
        self.bayesian_information_criterion = bayesian_information_criterion
        self.raw_estimation_results = raw_estimation_results


class TestMultiObjectives(unittest.TestCase):
    def test_loglikelihood_dimension_valid(self):
        res = DummyResults(final_log_likelihood=-123.45, number_of_parameters=4)
        result = loglikelihood_dimension(res)
        self.assertEqual(result, [123.45, 4])

    def test_loglikelihood_dimension_invalid(self):
        res = DummyResults(raw_estimation_results=None)
        result = loglikelihood_dimension(res)
        max_float = float(np.finfo(np.float32).max)
        self.assertEqual(result, [max_float, max_float])

    def test_aic_bic_dimension_valid(self):
        res = DummyResults(
            akaike_information_criterion=456.78,
            bayesian_information_criterion=478.90,
            number_of_parameters=5,
        )
        result = aic_bic_dimension(res)
        self.assertEqual(result, [456.78, 478.90, 5])

    def test_aic_bic_dimension_invalid(self):
        res = DummyResults(raw_estimation_results=None)
        result = aic_bic_dimension(res)
        max_float = float(np.finfo(np.float32).max)
        self.assertEqual(result, [max_float, max_float, max_float])

    def test_deprecated_AIC_BIC_dimension(self):
        res = DummyResults()
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            AIC_BIC_dimension(res)
            self.assertTrue(
                any("deprecated" in str(warning.message).lower() for warning in w)
            )


if __name__ == "__main__":
    unittest.main()



"""Tests of the objects for visiting arithmetic expressions

Michel Bierlaire
15.04.2025 16:56
"""

import unittest
from typing import Any

from biogeme.expressions import Expression
from biogeme.expressions.collectors import ExpressionCollector
from biogeme.expressions.visitor import ExpressionVisitor


class DummyExpression(Expression):
    """A dummy expression used for testing purposes."""

    def __init__(self, name: str, children=None):
        self.name = name
        self._children = children or []
        super().__init__()

    def get_children(self) -> list[Expression]:
        return self._children


class TestExpressionWalkerVisitor(unittest.TestCase):
    def setUp(self):
        self.leaf = DummyExpression("leaf")
        self.child = DummyExpression("child", [self.leaf])
        self.root = DummyExpression("root", [self.child])

    def test_expression_walker_collects_results(self):
        walker = ExpressionCollector()

        @walker.register(DummyExpression)
        def handler(expr: DummyExpression, context: Any) -> list[str]:
            return [f"handled: {expr.name}"]

        results = walker.walk(self.root)
        expected = ["handled: root", "handled: child", "handled: leaf"]
        self.assertEqual(results, expected)

    def test_expression_walker_invalid_handler_output(self):
        walker = ExpressionCollector()

        @walker.register(DummyExpression)
        def handler(expr: DummyExpression, context: Any):
            return "invalid"

        with self.assertRaises(TypeError):
            walker.walk(self.root)

    def test_expression_visitor_executes_handlers(self):
        visitor = ExpressionVisitor()
        visited = []

        @visitor.register(DummyExpression)
        def handler(expr: DummyExpression, context: Any) -> None:
            visited.append(expr.name)

        visitor.visit(self.root)
        self.assertEqual(visited, ["root", "child", "leaf"])

    def test_expression_visitor_invalid_handler_output(self):
        visitor = ExpressionVisitor()

        @visitor.register(DummyExpression)
        def handler(expr: DummyExpression, context: Any) -> str:
            return "not None"

        with self.assertRaises(TypeError):
            visitor.visit(self.root)


class TestExpressionVisitor(unittest.TestCase):
    def setUp(self):
        # Create a simple tree: root -> [child -> [leaf]]
        self.leaf = DummyExpression("leaf")
        self.child = DummyExpression("child", [self.leaf])
        self.root = DummyExpression("root", [self.child])

    def test_basic_traversal(self):
        visitor = ExpressionVisitor()
        visited = []

        @visitor.register(DummyExpression)
        def handle(expr: DummyExpression, context: Any) -> None:
            visited.append(expr.name)

        visitor.visit(self.root)
        self.assertEqual(visited, ["root", "child", "leaf"])

    def test_no_handler(self):
        # Test that traversal works fine with no registered handlers
        visitor = ExpressionVisitor()
        try:
            visitor.visit(self.root)
        except Exception as e:
            self.fail(f"Visitor raised an exception unexpectedly: {e}")

    def test_handler_must_return_none(self):
        visitor = ExpressionVisitor()

        @visitor.register(DummyExpression)
        def invalid_handler(expr: DummyExpression, context: Any) -> str:
            return "not none"

        with self.assertRaises(TypeError) as context:
            visitor.visit(self.root)
        self.assertIn("must return None", str(context.exception))

    def test_partial_handler_registration(self):
        visitor = ExpressionVisitor()
        visited = []

        class AnotherExpression(Expression):
            def get_children(self) -> list[Expression]:
                return []

        @visitor.register(DummyExpression)
        def dummy_handler(expr: DummyExpression, context: Any) -> None:
            visited.append(expr.name)

        mixed = DummyExpression("dummy", [AnotherExpression()])
        visitor.visit(mixed)
        self.assertEqual(visited, ["dummy"])  # Only DummyExpression triggers handler


if __name__ == "__main__":
    unittest.main()



"""

Numerical integration
=====================

Calculation of a mixtures of logit models where the integral is
calculated using numerical integration.

:author: Michel Bierlaire, EPFL
:date: Thu Apr 13 20:51:32 2023
"""

import biogeme.biogeme as bio
import biogeme.distributions as dist
from biogeme.expressions import RandomVariable, Integrate
from biogeme import models

from swissmetro_one import (
    database,
    TRAIN_TT_SCALED,
    TRAIN_COST_SCALED,
    SM_TT_SCALED,
    SM_COST_SCALED,
    CAR_TT_SCALED,
    CAR_CO_SCALED,
    TRAIN_AV_SP,
    SM_AV,
    CAR_AV_SP,
    CHOICE,
)

# %%
# Parameters
ASC_CAR = 0.137
ASC_TRAIN = -0.402
ASC_SM = 0
B_TIME = -2.26
B_TIME_S = 1.66
B_COST = -1.29

# %%
# Define a random parameter, normally distributed,
# designed to be used for integration
omega = RandomVariable('omega')
density = dist.normalpdf(omega)
B_TIME_RND = B_TIME + B_TIME_S * omega

# %%
# Definition of the utility functions
V1 = ASC_TRAIN + B_TIME_RND * TRAIN_TT_SCALED + B_COST * TRAIN_COST_SCALED
V2 = ASC_SM + B_TIME_RND * SM_TT_SCALED + B_COST * SM_COST_SCALED
V3 = ASC_CAR + B_TIME_RND * CAR_TT_SCALED + B_COST * CAR_CO_SCALED

# %%
# Associate utility functions with the numbering of alternatives
V = {1: V1, 2: V2, 3: V3}

# %%
# Associate the availability conditions with the alternatives
av = {1: TRAIN_AV_SP, 2: SM_AV, 3: CAR_AV_SP}

# %%
# The choice model is a logit, with availability conditions
integrand = models.logit(V, av, CHOICE)
numericalI = Integrate(integrand * density, 'omega')

# %%
simulate = {'Numerical': numericalI}

# %%
biosim = bio.BIOGEME(database, simulate)

# %%
results = biosim.simulate(theBetaValues={})
results

# %%
print('Mixture of logit - numerical integration: ', results.iloc[0]['Numerical'])



"""

Ordinal logit model
===================

Example of an ordinal logit model.  This is just to illustrate the
syntax, as the data are not ordered.  But the example assume, for the
sake of it, that the alternatives are ordered as 1->2->3

:author: Michel Bierlaire, EPFL
:date: Mon Apr 10 12:15:28 2023

"""

import biogeme.biogeme_logging as blog
import biogeme.biogeme as bio
from biogeme.models import ordered_logit
from biogeme.expressions import Beta, log, Elem

# %%
# See the data processing script: :ref:`swissmetro_data`.
from swissmetro_data import (
    database,
    CHOICE,
    TRAIN_TT_SCALED,
    TRAIN_COST_SCALED,
)

logger = blog.get_screen_logger(level=blog.INFO)
logger.info('Example b18ordinal_logit.py')

# %%
# Parameters to be estimated
B_TIME = Beta('B_TIME', 0, None, None, 0)
B_COST = Beta('B_COST', 0, None, None, 0)

# %%
# Threshold parameters for the ordered logit.

# %%
# :math:`\tau_1 \leq 0`.
tau1 = Beta('tau1', -1, None, 0, 0)

# %%
# :math:`\delta_2 \geq 0`.
delta2 = Beta('delta2', 2, 0, None, 0)

# %%
# :math:`\tau_2 = \tau_1 + \delta_2`
tau2 = tau1 + delta2

# %%
#  Utility.
U = B_TIME * TRAIN_TT_SCALED + B_COST * TRAIN_COST_SCALED

# %%
# Associate each discrete indicator with an interval.
#
#   1. :math:`-\infty \to \tau_1`,
#   2. :math:`\tau_1 \to \tau_2`,
#   3. :math:`\tau_2 \to +\infty`.
the_proba = ordered_logit(
    continuous_value=U,
    list_of_discrete_values=[1, 2, 3],
    tau_parameter=tau1,
)

# %%
# Extract from the dict the formula associated with the observed choice.
the_chosen_proba = Elem(the_proba, CHOICE)

# %%
# Definition of the model. This is the contribution of each
# observation to the log likelihood function.
logprob = log(the_chosen_proba)

# %%
# Create the Biogeme object.
the_biogeme = bio.BIOGEME(database, logprob)
the_biogeme.modelName = 'b18ordinal_logit'

# %%
# Estimate the parameters
results = the_biogeme.estimate()

# %%
print(results.short_summary())

# %%
pandas_results = results.getEstimatedParameters()
pandas_results



import io
import unittest
from datetime import timedelta
from unittest.mock import MagicMock, mock_open, patch

import numpy as np

from biogeme.results_processing import RawEstimationResults
from biogeme.results_processing.recycle_pickle import (
    Beta,
    BiogemeUnpickler,
    RawResults,
    pickle_to_yaml,
    read_pickle_biogeme,
)


class TestRecyclePickle(unittest.TestCase):

    def setUp(self):
        # Dummy beta parameters
        beta = Beta()
        beta.name = 'ASC_CAR'
        beta.value = 1.23
        beta.lb = -10
        beta.ub = 10

        # Dummy results object
        self.raw_results = RawResults()
        self.raw_results.modelName = 'TestModel'
        self.raw_results.userNotes = 'Test note'
        self.raw_results.betas = [beta]
        self.raw_results.g = np.array([1.0])
        self.raw_results.H = np.array([[1.0]])
        self.raw_results.bhhh = np.array([[1.0]])
        self.raw_results.nullLogLike = -100.0
        self.raw_results.initLogLike = -120.0
        self.raw_results.logLike = -95.0
        self.raw_results.dataname = 'testdata'
        self.raw_results.sampleSize = 100
        self.raw_results.numberOfObservations = 100
        self.raw_results.monte_carlo = False
        self.raw_results.numberOfDraws = 0
        self.raw_results.typesOfDraws = {}
        self.raw_results.excludedData = 0
        self.raw_results.drawsProcessingTime = timedelta(seconds=2)
        self.raw_results.optimizationMessages = {'message': 'success'}
        self.raw_results.convergence = True
        self.raw_results.bootstrap = np.array([[1.1]])
        self.raw_results.bootstrap_time = timedelta(seconds=1)

    @patch('builtins.open', new_callable=mock_open)
    @patch('pickle.load')
    def test_read_pickle_biogeme_success(self, mock_pickle_load, mock_file):
        mock_pickle_load.return_value = self.raw_results
        result = read_pickle_biogeme('dummy.pkl')
        self.assertIsInstance(result, RawEstimationResults)
        self.assertEqual(result.model_name, 'TestModel')
        self.assertEqual(result.beta_names, ['ASC_CAR'])

    @patch('biogeme.results_processing.recycle_pickle.read_pickle_biogeme')
    @patch('biogeme.results_processing.recycle_pickle.serialize_to_yaml')
    def test_pickle_to_yaml(self, mock_serialize, mock_reader):
        mock_result = MagicMock(spec=RawEstimationResults)
        mock_reader.return_value = mock_result

        pickle_file = 'input.pkl'
        yaml_file = 'output.yaml'
        pickle_to_yaml(pickle_file, yaml_file)

        mock_reader.assert_called_once_with(filename=pickle_file)
        mock_serialize.assert_called_once_with(data=mock_result, filename=yaml_file)

    def test_biogeme_unpickler_custom_types(self):
        dummy_file = io.BytesIO(b"dummy")
        unpickler = BiogemeUnpickler(file=dummy_file)
        self.assertIs(unpickler.find_class('biogeme.results', 'beta'), Beta)
        self.assertIs(unpickler.find_class('biogeme.results', 'rawResults'), RawResults)

    def test_beta_class_fields(self):
        beta = Beta()
        self.assertIsNone(beta.name)
        self.assertIsNone(beta.stdErr)
        beta.name = 'B_COST'
        self.assertEqual(beta.name, 'B_COST')


if __name__ == '__main__':
    unittest.main()



""".. _plot_b22multiple_models_spec:

Specification of a catalog of models
====================================

Specification of the Catalog of expressions for the assisted
specification algorithm. Note that this script does not perform any
estimation. It is imported by other scripts:
:ref:`plot_b22multiple_models`, :ref:`plot_b22process_pareto`.

:author: Michel Bierlaire, EPFL
:date: Fri Jul 21 17:56:47 2023

"""
from biogeme import models
import biogeme.biogeme as bio
from biogeme.expressions import Beta, logzero
from biogeme.catalog import Catalog, segmentation_catalogs

# %%
# See the data processing script: :ref:`swissmetro_data`.
from swissmetro_data import (
    database,
    CHOICE,
    SM_AV,
    CAR_AV_SP,
    TRAIN_AV_SP,
    TRAIN_TT_SCALED,
    TRAIN_COST_SCALED,
    SM_TT_SCALED,
    SM_COST_SCALED,
    CAR_TT_SCALED,
    CAR_CO_SCALED,
    MALE,
    TRAIN_HE,
    SM_HE,
    LUGGAGE,
    GA,
)

# %%
# Parameters to be estimated.
ASC_CAR = Beta('ASC_CAR', 0, None, None, 0)
ASC_TRAIN = Beta('ASC_TRAIN', 0, None, None, 0)
ASC_CAR = Beta('ASC_CAR', 0, None, None, 0)
ASC_TRAIN = Beta('ASC_TRAIN', 0, None, None, 0)
B_TIME = Beta('B_TIME', 0, None, None, 0)
B_COST = Beta('B_COST', 0, None, None, 0)
B_HEADWAY = Beta('B_HEADWAY', 0, None, None, 0)

# %%
# Define segmentations
gender_segmentation = database.generate_segmentation(
    variable=MALE, mapping={0: 'female', 1: 'male'}
)

GA_segmentation = database.generate_segmentation(
    variable=GA, mapping={0: 'without_ga', 1: 'with_ga'}
)

luggage_segmentation = database.generate_segmentation(
    variable=LUGGAGE, mapping={0: 'no_lugg', 1: 'one_lugg', 3: 'several_lugg'}
)

ASC_CAR_catalog, ASC_TRAIN_catalog = segmentation_catalogs(
    generic_name='ASC',
    beta_parameters=[ASC_CAR, ASC_TRAIN],
    potential_segmentations=[
        gender_segmentation,
        luggage_segmentation,
        GA_segmentation,
    ],
    maximum_number=2,
)

# %%
# We define a catalog with two different specifications for headway.
TRAIN_HEADWAY_catalog = Catalog.from_dict(
    catalog_name='TRAIN_HEADWAY_catalog',
    dict_of_expressions={'without_headway': 0, 'with_headway': B_HEADWAY * TRAIN_HE},
)

SM_HEADWAY_catalog = Catalog.from_dict(
    catalog_name='SM_HEADWAY_catalog',
    dict_of_expressions={'without_headway': 0, 'with_headway': B_HEADWAY * SM_HE},
    controlled_by=TRAIN_HEADWAY_catalog.controlled_by,
)


# %%
# Parameter for Box-Cox transforms.
ell_TT = Beta('lambda_TT', 1, None, 10, 0)

# %%
# Non linear specification for travel time.
TRAIN_TT_catalog = Catalog.from_dict(
    catalog_name='TRAIN_TT_catalog',
    dict_of_expressions={
        'linear': TRAIN_TT_SCALED,
        'log': logzero(TRAIN_TT_SCALED),
        'sqrt': TRAIN_TT_SCALED**0.5,
        'piecewise_1': models.piecewiseFormula(TRAIN_TT_SCALED, [0, 0.1, None]),
        'piecewise_2': models.piecewiseFormula(TRAIN_TT_SCALED, [0, 0.25, None]),
        'boxcox': models.boxcox(TRAIN_TT_SCALED, ell_TT),
    },
)

SM_TT_catalog = Catalog.from_dict(
    catalog_name='SM_TT_catalog',
    dict_of_expressions={
        'linear': SM_TT_SCALED,
        'log': logzero(SM_TT_SCALED),
        'sqrt': SM_TT_SCALED**0.5,
        'piecewise_1': models.piecewiseFormula(SM_TT_SCALED, [0, 0.1, None]),
        'piecewise_2': models.piecewiseFormula(SM_TT_SCALED, [0, 0.25, None]),
        'boxcox': models.boxcox(SM_TT_SCALED, ell_TT),
    },
    controlled_by=TRAIN_TT_catalog.controlled_by,
)

CAR_TT_catalog = Catalog.from_dict(
    catalog_name='CAR_TT_catalog',
    dict_of_expressions={
        'linear': CAR_TT_SCALED,
        'log': logzero(CAR_TT_SCALED),
        'sqrt': CAR_TT_SCALED**0.5,
        'piecewise_1': models.piecewiseFormula(CAR_TT_SCALED, [0, 0.1, None]),
        'piecewise_2': models.piecewiseFormula(CAR_TT_SCALED, [0, 0.25, None]),
        'boxcox': models.boxcox(CAR_TT_SCALED, ell_TT),
    },
    controlled_by=TRAIN_TT_catalog.controlled_by,
)

# %%
# Parameter for Box-Cox transforms.
ell_COST = Beta('lambda_COST', 1, None, 10, 0)

# %%
# Nonlinear transformations for travel cost.
TRAIN_COST_catalog = Catalog.from_dict(
    catalog_name='TRAIN_COST_catalog',
    dict_of_expressions={
        'linear': TRAIN_COST_SCALED,
        'log': logzero(TRAIN_COST_SCALED),
        'sqrt': TRAIN_COST_SCALED**0.5,
        'piecewise_1': models.piecewiseFormula(TRAIN_COST_SCALED, [0, 0.1, None]),
        'piecewise_2': models.piecewiseFormula(TRAIN_COST_SCALED, [0, 0.25, None]),
        'boxcox': models.boxcox(TRAIN_COST_SCALED, ell_COST),
    },
)

SM_COST_catalog = Catalog.from_dict(
    catalog_name='SM_COST_catalog',
    dict_of_expressions={
        'linear': SM_COST_SCALED,
        'log': logzero(SM_COST_SCALED),
        'sqrt': SM_COST_SCALED**0.5,
        'piecewise_1': models.piecewiseFormula(SM_COST_SCALED, [0, 0.1, None]),
        'piecewise_2': models.piecewiseFormula(SM_COST_SCALED, [0, 0.25, None]),
        'boxcox': models.boxcox(SM_COST_SCALED, ell_COST),
    },
    controlled_by=TRAIN_COST_catalog.controlled_by,
)

CAR_COST_catalog = Catalog.from_dict(
    catalog_name='CAR_COST_catalog',
    dict_of_expressions={
        'linear': CAR_CO_SCALED,
        'log': logzero(CAR_CO_SCALED),
        'sqrt': CAR_CO_SCALED**0.5,
        'piecewise_1': models.piecewiseFormula(CAR_CO_SCALED, [0, 0.1, None]),
        'piecewise_2': models.piecewiseFormula(CAR_CO_SCALED, [0, 0.25, None]),
        'boxcox': models.boxcox(CAR_CO_SCALED, ell_COST),
    },
    controlled_by=TRAIN_COST_catalog.controlled_by,
)

# %%
# Definition of the utility functions
V1 = (
    ASC_TRAIN_catalog
    + B_TIME * TRAIN_TT_catalog
    + B_COST * TRAIN_COST_catalog
    + TRAIN_HEADWAY_catalog
)
V2 = B_TIME * SM_TT_catalog + B_COST * SM_COST_catalog + SM_HEADWAY_catalog

V3 = ASC_CAR_catalog + B_TIME * CAR_TT_catalog + B_COST * CAR_COST_catalog

# %%
# Associate utility functions with the numbering of alternatives
V = {1: V1, 2: V2, 3: V3}

# %%
# Associate the availability conditions with the alternatives
av = {1: TRAIN_AV_SP, 2: SM_AV, 3: CAR_AV_SP}

# %%
# Definition of the model. This is the contribution of each
# observation to the log likelihood function.
logprob = models.loglogit(V, av, CHOICE)

# %%
the_biogeme = bio.BIOGEME(database, logprob)
the_biogeme.modelName = 'b22multiple_models'

# %%
PARETO_FILE_NAME = 'b22multiple_models.pareto'



"""Tests for binary expressions

Michel Bierlaire
Wed Mar 26 11:26:56 2025
"""

import unittest
import warnings

from biogeme.exceptions import BiogemeError
from biogeme.expressions import Numeric, bioMax, bioMin
from biogeme.expressions.binary_max import BinaryMax
from biogeme.expressions.binary_min import BinaryMin
from biogeme.expressions.divide import Divide
from biogeme.expressions.logical_and import And
from biogeme.expressions.logical_or import Or
from biogeme.expressions.minus import Minus
from biogeme.expressions.plus import Plus
from biogeme.expressions.power import Power
from biogeme.expressions.times import Times


class TestComprehensiveBinaryExpressions(unittest.TestCase):
    def setUp(self):
        self.a = Numeric(3)
        self.b = Numeric(2)
        self.zero = Numeric(0)

    def test_expressions(self):
        operators = [
            ('Plus', Plus, '+', lambda x, y: x + y),
            ('Minus', Minus, '-', lambda x, y: x - y),
            ('Times', Times, '*', lambda x, y: x * y),
            ('Divide', Divide, '/', lambda x, y: x / y),
            ('Power', Power, '**', lambda x, y: x**y),
            ('BinaryMin', BinaryMin, 'BinaryMin', min),
            ('BinaryMax', BinaryMax, 'BinaryMax', max),
            ('And', And, 'and', lambda x, y: 1 if x and y else 0),
            ('Or', Or, 'or', lambda x, y: 1 if x or y else 0),
        ]
        for name, op_class, op_symbol, func in operators:
            with self.subTest(operator=name):
                expr = op_class(self.a, self.b)
                expected = func(3, 2)
                # Test get_value
                self.assertEqual(expr.get_value(), expected)
                # Test __str__ contains operator symbol or name
                self.assertIn(str(op_symbol), str(expr))
                # Test __repr__ contains operator name for BinaryMin/Max, else operator symbol
                if name in ['BinaryMin', 'BinaryMax']:
                    self.assertIn(name, repr(expr))
                else:
                    self.assertIn(op_symbol, repr(expr))
                # Test JAX function correctness
                jax_fn = expr.recursive_construct_jax_function(numerically_safe=False)
                self.assertAlmostEqual(
                    float(jax_fn(None, None, None, None)), float(expected), places=5
                )

    def test_divide_by_zero(self):
        expr = Divide(self.a, self.zero)
        with self.assertRaises(BiogemeError):
            expr.get_value()

    def test_biomin_deprecation(self):
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            _ = bioMin(self.a, self.b)
            self.assertTrue(
                any(issubclass(item.category, DeprecationWarning) for item in w)
            )

    def test_biomax_deprecation(self):
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            _ = bioMax(self.a, self.b)
            self.assertTrue(
                any(issubclass(item.category, DeprecationWarning) for item in w)
            )


if __name__ == '__main__':
    import unittest

    unittest.main()



from .factory import DrawFactory
from .generators import (
    get_antithetic,
    get_halton_draws,
    get_latin_hypercube_draws,
    get_normal_wichura_draws,
    get_uniform,
)
from .management import DrawsManagement
from .native_draws import (
    RandomNumberGeneratorTuple,
    description_of_native_draws,
    native_random_number_generators,
)
